#include "Level.h"
#include <allegro.h>
#include <stdio.h>
#include <string.h>
#include "AllegroShortcuts.h"


// Create a new level of size (w, h)
Level::Level (const int w, const int h) 
{
    number = 0;
    title = NULL;
    filename = NULL;
    Initialize (w, h);
}



// Create a new level with size and contents equal to other.
Level::Level (Level * other)
{
    width = other->GetWidth();
    height = other->GetHeight();

    Initialize  (width, height);
    Copy        (other);
}



Level::~Level ()
{
    if (title != NULL) delete [] title;
    if (filename != NULL) delete [] filename;
    Clear();
}



// Set the size of the level and allocate memory for it
void Level::Initialize (const int w, const int h)
{
    int i, j;

    width = w;
    height = h;

    tiles = new int * [height];
    for (i = 0; i < height; i++)
    {
        // Create a new row of tiles and initialize them all to empty
        tiles[i] = new int[width];
        for (j = 0; j < width; j++)
        {
            tiles[i][j] = 0;
        }
    }
}

 

// Clear the memory allocated to the level
void Level::Clear (void)
{
    int i;

    for (i = 0; i < height; i++) delete [] tiles[i];
    delete [] tiles;
}



// Makes this level a copy of other. If one level is smaller than the other,
// copies the smaller area.
void Level::Copy (Level * other)
{
    int x, y, useWidth, useHeight;

    // Determine the size of the area to copy (use the smaller of each of the
    // two levels' widths and heights)
    if (width < other->GetWidth())
    {
        useWidth = width;
    }
    else
    {
        useWidth = other->GetWidth();
    }

    if (height < other->GetHeight())
    {
        useHeight = height;
    }
    else
    {
        useHeight = other->GetHeight();
    }

    // Copy the contents of the other level to this one
    for (x = 0; x < useWidth; x++)
    {
        for (y = 0; y < useHeight; y++)
        {
            SetTile (x, y, other->GetTile(x, y));
        }    
    }    
}



// Sets all the tiles in this level to the specified tile
void Level::ClearTilesTo (const int which)
{
    int x, y;

    for (y = 0; y < height; y++)
    {
        for (x = 0; x < width; x++)
        {
            SetTile (x, y, which);
        }
    }
}



void Level::SetTitle (const char * str)
{
    // Clear the memory for the previous title if necessary
    if (title != NULL) delete [] title;
    title = new char[strlen(str) + 1];
    strcpy (title, str);
}



void Level::SetFilename (const char * str)
{
    // Clear the memory for the previous filename if necessary
    if (filename != NULL) delete filename;
    filename = new char[strlen(str) + 1];
    strcpy (filename, str);
}



void Level::SetNumber (const int num)   { number = num; }



void Level::SetTile (const int x, const int y, const int tile)
{
    tiles[y][x] = tile;
}



char * Level::GetTitle      (void)                      { return title; }
char * Level::GetFilename   (void)                      { return filename; }
int Level::GetNumber        (void)                      { return number; }
int Level::GetWidth         (void)                      { return width; }
int Level::GetHeight        (void)                      { return height; }



// Returns the tile at (x, y), or -1 if (x, y) is out of bounds
int Level::GetTile (const int x, const int y)
{
    if (x < 0 || y < 0 || x >= width || y >= height)
    {
        return -1;
    }
    else
    {
        return tiles[y][x];
    }
}



// Save a level to the filename specified by this Level's filename field.
// If an error occurs, display an error message.
void Level::Save (void)
{
    // Assemble filename before passing in here: filename = Levels/levelX.dat

    FILE * out;
    int x, y,
        tile,
        type;

    out = fopen (filename, "wb");
    if (out == NULL)
    {
        allegro_message ("Save failed!\nCould not open \"%s\" for output.",
            filename);
        return;
    }

    // No use for "type" variable - remove it from the level standard!

    fwrite (GetTitle(), sizeof(char) * LEVEL_TITLE_LENGTH,  1, out);
    fwrite (&width,     sizeof(int),                        1, out);
    fwrite (&height,    sizeof(int),                        1, out);
    fwrite (&type,      sizeof(int),                        1, out);

    for (y = 0; y < height; y++)
    {
        for (x = 0; x < width; x++)
        {
            tile = GetTile (x, y);
            fwrite (&tile, sizeof(int), 1, out);
        }
    }

    fclose (out);
}



// Load a level from the filename specified by this Level's filename field.
// If an error occurs, display an error message.
void Level::Load (void)
{
    // Assemble filename before passing in here: filename = Levels/levelX.dat

    FILE * in;
    char str [LEVEL_TITLE_LENGTH];
    int x, y,
        newWidth, newHeight,
        tile,
        type;

    in = fopen (filename, "rb");
    if (in == NULL)
    {
        allegro_message ("Load failed!\nCould not open \"%s\" for input.",
            filename);
        return;
    }

    fread (str,         sizeof(char) * LEVEL_TITLE_LENGTH,  1, in);
    fread (&newWidth,   sizeof(int),                        1, in);
    fread (&newHeight,  sizeof(int),                        1, in);
    fread (&type,       sizeof(int),                        1, in);

    Clear       ();
    Initialize  (newWidth, newHeight);
    SetTitle    (str);

    for (y = 0; y < height; y++)
    {
        for (x = 0; x < width; x++)
        {
            fread (&tile, sizeof(int), 1, in);
            SetTile (x, y, tile);
        }
    }

    fclose (in);
}



// Sets the size of the level. If the new size is larger than its current size,
// fills the new cells with emptyTile.
void Level::Resize (const int newWidth, const int newHeight,
    const int emptyTile)
{
    int x, y,
        prevWidth = width, prevHeight = height;

    // Make a copy of this level
    Level * tmp = new Level (this);

    // Resize the level to the specified new width and height.
    // Then restore the contents of the level (that we copied into tmp).
    Clear       ();
    Initialize  (newWidth, newHeight);
    Copy        (tmp);

    // If the level has grown in either dimension, fill the new cells
    for (x = prevWidth; x < newWidth; x++)
    {
        for (y = 0; y < newHeight; y++)
        {
            SetTile (x, y, emptyTile);
        }
    }
    for (y = prevHeight; y < newHeight; y++)
    {                
        for (x = 0; x < newWidth; x++)
        {
            SetTile (x, y, emptyTile);
        }
    }
}
