#include "UserInterface.h"
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "AllegroShortcuts.h"

#define MAX_STR_LEN 500
#define NUM_PIECES  7

UserInterface::UserInterface (
    const int screenWidth,
    const int screenHeight,
    const int colorDepth,
    const int gfxMode,
    const int levelWidth,
    const int levelHeight,
    const int tileWidth_,
    const int tileHeight_
    )
{
    int windowXPixel = screenWidth / 2 - (tileWidth_ * levelWidth / 2),
        windowYPixel = screenHeight / 2 - (tileHeight_ * levelHeight / 2);

    // Create and initialize the objects in order
    graphics =  new Graphics (screenWidth, screenHeight, colorDepth, gfxMode);
    keyboard =  new Keyboard ();
    mouse =     new Mouse    ();
    board =     new Board    (levelWidth, levelHeight);
    window =    new Window   (levelWidth, levelHeight, windowXPixel,
                                windowYPixel, levelWidth, levelHeight);    
    timer =     new Timer    ();
    
    score = 0;
    level = 1;
    showNext = true;
    lines = 0;
    nextPiece = rand() % NUM_PIECES;

    timer->SetLength (LEVEL_FALL_BASE + LEVEL_FALL_INC * (level - 1));

    for (int i = 0; i < NUM_PIECES; i++)
    {
        pieces[i].Initialize (4, 4, 4, TETRIS_PIECES[i]);
    }
    
    tileWidth = tileWidth_;
    tileHeight = tileHeight_;

    LoadBackground ();

    board->Clear            ();
    graphics->ShowMouse     ();

    CreatePiece ();
    PieceToBoard (false);

    DrawWindow ();
    graphics->Refresh ();

    allegro_message ("Welcome to FreeTetris!\n"
        "Created by Daniel Tenedorio.\n"
        "Email: daniel.tenedorio@gmail.com.\n\n"
        "Left arrow key: Move left\n"
        "Right arrow key: Move right\n"
        "Down arrow key: Drop piece faster\n"
        "Up arrow key: Rotate piece\n"
        "Esc key: Quit game\n\n"
        "This game does not break any patents or copyrights.\n"
        "By clicking OK, you agree to one of twelve licenses,\n"
        "selected at random. Thank you for playing.");
}



void UserInterface::Loop (void)
{
    while (true)
    {
        if (timer->HasTriggered ())
        {
            if (MovePiece (0, 1)) 
            {
                timer->Reset ();
            }
        }

        graphics->Refresh ();
        graphics->YieldCPU ();
        
        RespondToKeyboard ();
    }
}



void UserInterface::RespondToKeyboard (void)
{
    static Timer moveTimer (MOVE_DELAY);
    static Timer dropTimer (DROP_DELAY);
    bool delayed = false;

    keyboard->ScanInput ();

    // If the Esc key is pressed, exit
    if (keyboard->GetKeyDown (ESC_KEY)) 
    {
        allegro_exit();
        exit(0);
    }

    // If the F1 key is pressed, display help 
    if (keyboard->GetKeyNewlyReleased (F1_KEY))
    {
        DisplayCommands();
    }

    // If any of the arrow keys are pressed, move the current piece
    if (keyboard->GetKeyNewlyDown (UP_KEY))
    {
        RotatePiece ();
    }
    if (keyboard->GetKeyDown (DOWN_KEY))
    {
        if (dropTimer.HasTriggered())
        {
            MovePiece (0, 1);
            score += DROP_PIECE_SCORE;
            dropTimer.SetLength (DROP_DELAY);
            dropTimer.Reset ();
        }
    }
    if (keyboard->GetKeyDown (LEFT_KEY))
    {
        if (moveTimer.HasTriggered())
        {
            MovePiece (-1, 0); 
            moveTimer.SetLength (MOVE_DELAY);
            moveTimer.Reset ();
        }
    }
    if (keyboard->GetKeyDown (RIGHT_KEY))
    {
        if (moveTimer.HasTriggered())
        {
            MovePiece (1, 0);
            moveTimer.SetLength (MOVE_DELAY);
            moveTimer.Reset ();
        }
    }
}



// Selects the current piece and initializes it
void UserInterface::CreatePiece (void)
{
    curPiece = nextPiece;
    nextPiece = rand() % NUM_PIECES;
    pieces[curPiece].Reset ();
    pieces[nextPiece].Reset ();
    pieces[curPiece].SetLocation (board->GetWidth () / 2 - 2, -2);




    if (CheckCollisions (0, 0)) GameOver();
}



void UserInterface::GameOver (void)
{
    exit (0); 
}



// Sets the cells of the board to the current piece.
// If draw is true, draws the piece; otherwise, erases it.
void UserInterface::PieceToBoard (const bool draw)
{
    for (int y = 0; y < pieces[curPiece].GetHeight(); y++)
    {
        for (int x = 0; x < pieces[curPiece].GetWidth(); x++)
        {
            if (pieces[curPiece].GetCell(x, y))
            {
                board->SetCell (x + pieces[curPiece].GetXCell(),
                    y + pieces[curPiece].GetYCell(), draw ? curPiece + 1 : 0);
            }
        }
    }
}



// Rotates the piece 90 degrees clockwise. If the newly rotated piece collides
// with a wall or another piece, move the piece to the left or right so it does
// not.
void UserInterface::RotatePiece (void)
{
    PieceToBoard (false);
    pieces[curPiece].Rotate (1);
    if (CheckCollisions(0, 0)) pieces[curPiece].Rotate (-1);
    PieceToBoard (true);
    DrawWindow ();
}



// Move the piece (x, y) cells, unless it collides with a wall or another piece.
// Return true if no collision occurred, or false if not.
bool UserInterface::MovePiece (const int x, const int y)
{
    PieceToBoard (false);           // Erase the piece to check for collisions
    if (CheckCollisions (x, y))
    {
        PieceToBoard (true);
        if (y > 0)
        {
            CreatePiece ();  // If the piece has hit botton, make new one
            DrawWindow ();
            CheckForLines ();
        }
        return false;
    }
    
    pieces[curPiece].IncLocation (x, y);        // Change the piece's location
    PieceToBoard (true);                        // Paste it back on the board
    DrawWindow ();

    return true;                                // No collisions occurred
}



// Check to see if there are any full lines; if so, phase them out
void UserInterface::CheckForLines (void)
{
    int x, y, y2, numFoundLines = 0;
    vector<int> foundLines;
    bool foundOneLine;
    Timer tempTimer;

    foundLines.clear();

    // Check for newly filled lines
    for (y = 0; y < board->GetHeight(); y++)
    {
        foundOneLine = true;
        for (x = 0; x < board->GetWidth(); x++)
        {
            if (board->GetCell (x, y) == 0)
            {
                foundOneLine = false;
            }
        }

        if (foundOneLine)
        {
            foundLines.push_back (y);
            score += ERASE_LINE_SCORE;
            lines++;
            
            // Advance to the next level
            if (lines > LEVEL_ADVANCE_LINES)
            {
                lines -= 10;
                level++;
                timer->SetLength (LEVEL_FALL_BASE - LEVEL_FALL_INC
                    * (level - 1));
                LoadBackground (); 
            }
        }
    }
    
    if (foundLines.size() == 0) return;

    // Blink the filled row(s) out
    for (int counter = 0; counter < NUM_BLINKS; counter++)
    {
        for (int whichCell = 0; whichCell < 2; whichCell++)
        {
            for (int lineCounter = 0; lineCounter < foundLines.size();
                lineCounter++)
            {
                y = foundLines[lineCounter];
                for (x = 0; x < board->GetWidth(); x++)
                {
                    board->SetCell (x, y, whichCell == 0 ? NUM_PIECES + 1 : 0);
                }
            }
            DrawWindow ();
            graphics->Refresh();
            tempTimer.Delay (BLINK_DELAY);
        }
    }

    // Shift all the cells above the row down one cell
    for (int lineCounter = 0; lineCounter < foundLines.size(); lineCounter++)
    {
        y = foundLines[lineCounter];
        for (y2 = y; y2 > 0; y2--)
        {
            for (x = 0; x < board->GetWidth(); x++)
            {
                board->SetCell (x, y2, board->GetCell (x, y2 - 1));
            }
        }
    }
}



// Returns whether or not any pieces and/or board edges are in the way if we
// wanted to move the current piece (x, y) cells.
bool UserInterface::CheckCollisions (const int x, const int y)
{
    int xCell, yCell, newXCell, newYCell;

    // Iterate through each of the 16 cells in our 4x4 piece. If a cell is
    // filled, check to see if the corresponding spot on the board is empty,
    // and make sure the spot is within the bounds of the board. 
    for (xCell = 0; xCell < 4; xCell++)
    {
        for (yCell = 0; yCell < 4; yCell++)
        {
            if (pieces[curPiece].GetCell (xCell, yCell))
            {
                newXCell = xCell + pieces[curPiece].GetXCell() + x;
                newYCell = yCell + pieces[curPiece].GetYCell() + y;

                if (newXCell < 0 || newXCell >= board->GetWidth () ||
                    newYCell >= board->GetHeight () ||
                    board->GetCell (newXCell, newYCell) != 0)
                {
                    return true;
                }
            }
        }
    }

    return false;    // No collision occurred
}



// Draws a cell at cell location (xCell, yCell).
void UserInterface::SetCell (const int xCell, const int yCell, const int cell)
{
    board->SetCell (xCell, yCell, cell);
}



// Draws a cell at (xPixel, yPixel)
void UserInterface::DrawCell (const int xPixel, const int yPixel, const int
    whichCell)
{
    Vector3 col;

    // Determine the color of the block to draw
    switch (whichCell)
    {
    // case 0:  col.SetXYZ (85, 85, 85);       break;  // Dark gray
    case 0:  col.SetXYZ (0, 0, 0);          break;  // Black
    case 1:  col.SetXYZ (255, 40, 40);      break;  // Red
    case 3:  col.SetXYZ (0, 255, 0);        break;  // Green
    case 2:  col.SetXYZ (40, 40, 255);      break;  // Blue
    case 4:  col.SetXYZ (0, 230, 230);      break;  // Yellow
    case 5:  col.SetXYZ (255, 0, 255);      break;  // Purple
    case 6:  col.SetXYZ (171, 171, 171);    break;  // Light gray
    case 7:  col.SetXYZ (255, 255, 0);      break;  // Yellow
    default: col.SetXYZ (255, 255, 255);    break;  // White 
    }
    
    // Draw it
    graphics->SetColor (col);
    graphics->FillRectangle (xPixel, yPixel, tileWidth, tileHeight);
}



void UserInterface::DrawWindow (void)
{
    char str[256];
    int xCell, yCell, whichCell, xPixel, yPixel;

    graphics->HideMouse ();

    // Display the background image
    graphics->DrawBackground ();

    // Draw the window
    for (yCell = 0; yCell < window->GetHeight(); yCell++)
    {
        for (xCell = 0; xCell < window->GetWidth(); xCell++)
        {   
            xPixel = window->GetStartXPixel() + (xCell) * tileWidth;
            yPixel = window->GetStartYPixel() + (yCell) * tileHeight;
            whichCell = board->GetCell (xCell, yCell);
            DrawCell (xPixel, yPixel, whichCell);
        }
    }

    // board->Print(); 

    // Draw the area behind the level/score
    // graphics->SetColor (85.0, 85.0, 85.0);  // Dark gray
    graphics->SetColor (0.0, 0.0, 0.0);      // Black
    graphics->FillRectangle (LEVEL_X - 20, LEVEL_Y - 20,
        40 + LEVEL_WIDTH, 40 + text_height (font) * 5);
   
    // Display the current level/score 
    graphics->SetColors (255.0, 255.0, 255.0, 105.0, 105.0, 105.0, 0.0, 0.0,
        0.0);
    sprintf (str, "Level %ld", level);
    graphics->LayeredTransText (LEVEL_X + 15, LEVEL_Y, str);
    sprintf (str, "Score %ld", score);
    graphics->LayeredTransText (LEVEL_X + 15, LEVEL_Y + text_height (font) * 2,
        str);
    sprintf (str, "Lines %ld", lines + ((level - 1) * 10));
    graphics->LayeredTransText (LEVEL_X + 15, LEVEL_Y + text_height (font) * 4,
        str);

    // Draw the area behind the next piece
    // graphics->SetColor (85.0, 85.0, 85.0);     // Dark gray
    graphics->SetColor (0.0, 0.0, 0.0);         // Black
    graphics->FillRectangle (NEXT_X - 20, NEXT_Y - 20,
        40 + NEXT_WIDTH, 40 + 20 + tileHeight * 2);
   
    // Display the next piece
    /*
    graphics->SetColors (255.0, 255.0, 255.0, 105.0, 105.0, 105.0, 0.0, 0.0,
        0.0);
    sprintf (str, "Next");
    graphics->LayeredTransText (NEXT_X + NEXT_WIDTH / 2 -
        text_length (font, "Next") / 2, NEXT_Y, str);
    */

    for (yCell = 1; yCell < 4; yCell++)
    {
        for (xCell = 0; xCell < 4; xCell++)
        {
            if (pieces[nextPiece].GetNonRotatedCell (xCell, yCell))
            {
                DrawCell (NEXT_X + xCell * tileWidth, NEXT_Y + (yCell - 1) *
                    tileHeight + 10, nextPiece + 1);
            }
        }
    }

    graphics->ShowMouse ();
}



// Shows the user the possible commands using an allegro dialog.
void UserInterface::DisplayCommands (void)
{
    allegro_message (
        "DANNY'S SHAMELESS TETRIS CLONE\n"
        "\n"
        "Arrow keys: Move current piece\n"
        "Space / Up arrow: Rotate current piece\n"
        "N: Toggle next piece\n"
        "F1: View this help\n"
        "Esc: Exit\n");
}



void UserInterface::LoadBackground (void)
{
    char str [2000];
    char shortStr [5] = "0";

    strcpy (str, BACKGROUND_IMAGE_FILENAME_BASE);
    shortStr[0] = '0' + level;
    strcat (str, shortStr);
    strcat (str, ".bmp");

    if (!graphics->LoadBackground (str))
    {
        allegro_message (BACKGROUND_MISSING_ERR);
        exit (1);
    }
}



